"""Модуль архиватора виртуальных машин Qemu/KVM."""

import os
import logging
import subprocess
import tempfile
import shutil
import tarfile
from datetime import date

from msbackup.backend_base import Base as BaseBackend
from msbackup.backend_kvm_snapshot import Snapshot


logger = logging.getLogger("msbackup")


def get_backend_kwargs(params):
    """Подготовка параметров для режима 'kvm'."""
    kwargs = BaseBackend.get_common_backend_kwargs(params)
    BaseBackend.get_param(params, kwargs, "pool")
    BaseBackend.get_param(params, kwargs, "filter")
    return kwargs


class Kvm(BaseBackend):
    """
    Архиватор виртуальных машин Qemu/KVM.

    Зависимости:
    - установлены пакеты "KVM":http://www.linux-kvm.org/page/Main_Page и
      "libvirt":http://libvirt.org/index.html;
    - "QEMU Guest agent":http://wiki.libvirt.org/page/Qemu_guest_agent
      установлен на каждой виртуальной машине;
    - образы дисков виртуальных машин в формате
      "qcow2":https://en.wikipedia.org/wiki/Qcow;
    - установлен "pigz":http://zlib.net/pigz;
    - установлен Python3 пакет "libvirt":https://libvirt.org/python.html.
    """

    SECTION = "Backend-KVM"
    LIBVIRT_CONNECTION = "qemu:///system"

    @classmethod
    def make_subparser(cls, subparsers):
        """Добавление раздела параметров командной строки для архиватора."""
        parser = subparsers.add_parser("kvm")
        parser.set_defaults(get_backend_kwargs=get_backend_kwargs)
        vm_filter = parser.add_mutually_exclusive_group()
        parser.add_argument(
            "--pool",
            default="default",
            help="Storage pool name to create snapshot.",
        )
        vm_filter.add_argument(
            "--active",
            dest="filter",
            action="store_const",
            const="active",
            help="Archive only active virtual machines (default).",
        )
        vm_filter.add_argument(
            "--inactive",
            dest="filter",
            action="store_const",
            const="inactive",
            help="Archive only inactive virtual machines.",
        )
        vm_filter.add_argument(
            "--all",
            dest="filter",
            action="store_const",
            const="all",
            help="Archive all virtual machines.",
        )
        parser.add_argument(
            "source",
            nargs="*",
            metavar="DOMAIN",
            help="Domain name of virtual machine.",
        )

    def __init__(self, config, *args, **kwargs):
        """Конструктор."""
        super().__init__(config, *args, **kwargs)
        import libvirt

        self.conn = libvirt.open(self.LIBVIRT_CONNECTION)
        self.cwd = None
        self.tmpdir = None
        self.tar_file = None
        self.pool = kwargs.get("pool") or config.get(
            self.SECTION, "POOL", fallback=None
        )
        self.filter = kwargs.get("filter") or config.get(
            self.SECTION, "FILTER", fallback="active"
        )

    def _backup(self, sources=None, **kwargs):
        """
        Архивация виртуальных машин Qemu/KVM.

        :param sources: Список имён виртуальных машин.
        :type sources: [str]
        :return: Количество ошибок.
        :rtype: int
        """
        import libvirt

        if sources is None:
            params = {}
            if self.filter == "active":
                params["flags"] = libvirt.VIR_CONNECT_LIST_DOMAINS_ACTIVE
            elif self.filter == "inactive":
                params["flags"] = libvirt.VIR_CONNECT_LIST_DOMAINS_INACTIVE
            domains = self.conn.listAllDomains(**params)
            sources = [dom.name() for dom in domains]
        error_count = 0
        for source in sources:
            output = self.outpath(source)
            logger.info("Backup domain: %s", source)
            try:
                self.archive(
                    source=source,
                    output=output,
                )
            except libvirt.libvirtError as ex:
                error_count += 1
                logger.error("libvirtError: %s", str(ex))
            except subprocess.CalledProcessError as ex:
                logger.error(str(ex))
                error_count += 1
        return error_count

    def _archive(self, source, output=None, **kwargs):
        """
        Упаковка состояния виртуальной машины в архив.

        :param source: Имя домена виртуальной машины.
        :type source: str
        :param output: Путь к файлу архива.
        :type output: str
        :return: Процесс упаковки файлов.
        :rtype: subprocess.Popen
        """
        snapshot = Snapshot(
            conn=self.conn,
            domain_name=source,
            pool=self.pool,
        )
        if snapshot.is_active() and snapshot.has_current_snapshot():
            raise RuntimeError('Domain "%s" has already a snapshot.' % source)
        ext, tar_mode = ".tar", "w"
        tar_name = source + ext
        self.tmpdir = tempfile.mkdtemp(dir=self.tmp_dir)
        self.cwd = os.getcwd()
        os.chdir(self.tmpdir)
        tar_path = os.path.join(self.tmpdir, tar_name)
        now_str = date.today().isoformat()
        with tarfile.open(tar_path, mode=tar_mode) as tar:
            xml_files = snapshot.dump_xml(path=self.tmpdir)
            for xml_file in xml_files:
                xml_file_name = os.path.basename(xml_file)
                tar.add(name=xml_file, arcname=os.path.join(now_str, xml_file_name))
                logger.debug("file %s added" % xml_file_name)
                os.remove(xml_file)
            snapshot.create_snapshot()
            logger.debug(
                'Adding image files for "%s" to archive "%s"',
                source,
                tar_path,
            )
            try:
                for disk, info in snapshot.disks.items():
                    if info["driver"] != "qcow2":
                        continue
                    src = info["file"]
                    file_name = os.path.basename(src)
                    logger.debug('Adding "%s" to archive "%s"', file_name, tar_path)
                    tar.add(
                        name=src,
                        arcname=os.path.join(now_str, file_name),
                    )
            except Exception as ex:
                logger.error("ERROR: %s", ex, exc_info=True)
                raise
            finally:
                if snapshot.is_active():
                    snapshot.do_blockcommit()
                    snapshot.remove_snapshot()
        logger.debug('Compressing "%s"', tar_name)
        if output is None:
            self.tar_file = open(tar_path, mode="rb")
            return self._compress_proc(
                in_stream=self.tar_file,
                out_stream=subprocess.PIPE,
            )
        else:
            with open(tar_path, mode="rb") as tar_file:
                with open(output, mode="wb") as out_file:
                    self._compress(in_stream=tar_file, out_stream=out_file)

    def _cleanup(self):
        """Удаление временных объектов."""
        if self.cwd is not None:
            os.chdir(self.cwd)
            self.cwd = None
        if self.tmpdir is not None:
            shutil.rmtree(self.tmpdir, ignore_errors=True)
            self.tmpdir = None
        if self.tar_file is not None:
            self.tar_file.close()
            self.tar_file = None
